<?php

namespace UnicaenCode\Service;

use Exception;
use UnicaenCode\Generator\GeneratorInterface;
use UnicaenCode\Service\Traits\ConfigServiceAwareTrait;
use UnicaenCode\Util;
use Zend\Stdlib\ArrayUtils;

/**
 *
 *
 * @author Laurent LÉCLUSE <laurent.lecluse at unicaen.fr>
 */
class CodeGeneratorService
{
    use ConfigServiceAwareTrait;


    /**
     * @var string
     */
    protected $outputDir;



    /**
     * @return string
     */
    public function getOutputDir()
    {
        if (!$this->outputDir) return $this->getServiceConfig()->getGeneratorOutputDir(); // par défaut celui de la config

        return $this->outputDir;
    }



    /**
     * @param string $outputDir
     *
     * @return CodeGenerator
     */
    public function setOutputDir($outputDir)
    {
        $this->outputDir = $outputDir;

        return $this;
    }



    /**
     * @param array $params
     *
     * Liste des paramètres à transmettre :
     *  - generator (string) Nom du générateur utilisé pour générer le fichier
     *  - write (?boolean) Définit si on écrit sur le disque dur ou non le fichier ainsi généré
     *  - echo (?boolean) Précise si on affiche ou non sur le navigateur le fichier ainsi généré
     *
     * Autres paramètres usuels à transmettre au générateur :
     *  - classname
     *
     *
     * @return array
     * @throws Exception
     */
    public function generate(array $params): array
    {
        /* Passage du générateur */
        $generator = $params['generator'] ?? null;

        if (is_string($generator)) {
            $generator = $this->getGenerator($generator);
        }
        if ($generator) {
            $params = $generator->generate($params);
        }

        foreach($params as $name => $value){
            if (is_array($value)){
                $subGenerator = $this->getGenerator(ucfirst($name));
                if ($subGenerator){
                    $this->subGenerate($subGenerator, $params, $value);
                }
            }
        }


        /* Génération du fichier final & rendu */
        $params = $this->makeCode($params);

        $this->echo($params);
        $this->write($params);

        return $params;
    }



    protected function subGenerate(GeneratorInterface $generator, array $parentParams, array $params)
    {
        $params['generator'] = $generator;

        // transfert des paramètres de sortie au sous-générateur
        if (!isset($params['echo']) && isset($parentParams['echo'])) $params['echo'] = $parentParams['echo'];
        if (!isset($params['echoExpanded']) && isset($parentParams['echoExpanded'])) $params['echoExpanded'] = $parentParams['echoExpanded'];
        if (!isset($params['write']) && isset($parentParams['write'])) $params['write'] = $parentParams['write'];

        $params = $generator->generateFromParent($parentParams, $params);

        return $this->generate($params);
    }



    /**
     *
     * @throws Exception
     */
    protected function makeCode(array $params)
    {
        $template = $params['template'] ?? null;

        if (!$template) return $params;

        $code = $this->getTemplateCode($template);

        foreach ($params as $param => $value) {
            if (!is_array($value) && !is_object($value)) {
                $code = $this->codeSetVariable($code, $param, $value);
            }
        }

        $code = trim($code); // éviter les espaces inutiles en début et fin de fichier

        /* On débarasse le code de ses caractères inutiles !! */
        $code = explode(PHP_EOL, $code);
        foreach ($code as $i => $line) {
            if (trim($line)) {
                $code[$i] = rtrim($line);
            } else {
                $code[$i] = '';
            }
        }
        $code = implode(PHP_EOL, $code);

        $params['code'] = $code;

        return $params;
    }



    public function getGenerator(string $generator): ?GeneratorInterface
    {
        $genClass = $generator;
        if (!class_exists($generator)) {
            $generator = 'UnicaenCode\\Generator\\' . $generator;
        }
        if (class_exists($generator)) {
            return new $generator;
        }

        return null;
    }



    public function getTemplateCode(string $template): string
    {
        $templateDirs = $this->getServiceConfig()->getTemplateDirs();
        $templateFile = null;
        foreach ($templateDirs as $templateDir) {
            $templateFile = $templateDir . '/' . $template . '.php';
            if (file_exists($templateFile)) break;
        }

        if (!$templateFile) {
            throw new Exception('Le modèle ' . $template . ' n\'a pas été trouvé dans les répertoires de modèles d\'UnicaenCode');
        }

        return file_get_contents($templateFile);
    }



    protected function codeSetVariable($code, $param, $value)
    {
        // Si c'est un tableau alors on concatène en string...
        if (is_array($param)) {
            $param = implode('', $param);
        }

        // application des paramètres simples
        $code = str_replace('<' . $param . '>', $value, $code);

        // Gestion des conditions
        $if = '<if ' . $param . '>';
        while (false !== ($begin = strpos($code, $if))) {
            $endIf = '<endif ' . $param . '>';

            $end = strpos($code, $endIf, $begin);

            if ($value) {
                $ifCode = ltrim(substr($code, $begin + strlen($if), $end - $begin - strlen($if)));
            } else {
                $ifCode = '';
            }

            $end += strlen($endIf);

            if (isset($code[$begin - 1]) && isset($code[$end])) {
                $b = trim($code[$begin - 1]);
                $e = trim($code[$end]);
                // ATTENTION : il n'y pas de support des retours chariot Windows (\n\r) car ils font 2 caractères ! ! !
                if ($b === '' && $e === '') $end++; // pour éviter les retours chariot ou espaces inutiles
            }

            $code = substr($code, 0, $begin) . $ifCode . substr($code, $end);
        }

        // Gestion des conditions
        $if = '<if ' . $param . ' notrim>';
        while (false !== ($begin = strpos($code, $if))) {
            $endIf = '<endif ' . $param . '>';

            $end = strpos($code, $endIf, $begin);

            if ($value) {
                $ifCode = substr($code, $begin + strlen($if), $end - $begin - strlen($if));
            } else {
                $ifCode = '';
            }

            $end += strlen($endIf);

            if (isset($code[$begin - 1]) && isset($code[$end])) {
                $b = trim($code[$begin - 1]);
                $e = trim($code[$end]);
                // ATTENTION : il n'y pas de support des retours chariot Windows (\n\r) car ils font 2 caractères ! ! !
                if ($b === '' && $e === '') $end++; // pour éviter les retours chariot ou espaces inutiles
            }

            $code = substr($code, 0, $begin) . $ifCode . substr($code, $end);
        }

        // restitution
        return $code;
    }



    /**
     * @param array $params
     *
     * @return $this
     */
    protected function echo(array $params): self
    {
        $echo = (bool)($params['echo'] ?? false);
        $expanded = (bool)($params['echoExpanded'] ?? false);
        $filename = $params['filename'] ?? 'Code source (fichier sans nom)';
        $code     = $params['code'] ?? null;

        if ($expanded) $echo = true;
        if (!$echo) return $this;

        $id = uniqid('bloc_code_');
        ?>
        <a role="button" data-toggle="collapse" href="#<?= $id ?>" aria-expanded="<?= $expanded ? 'true' : 'false' ?>" aria-controls="collapseExample">
            <?= $filename ?>
        </a><br/>
        <div class="collapse<?= $expanded ? ' in' : '' ?>" id="<?= $id ?>">
            <?php Util::highlight($code, 'php', true, ['show-line-numbers' => true]); ?>
        </div>
        <?php

        return $this;
    }



    /**
     * @param array $params
     *
     * @return $this
     */
    protected function write(array $params): self
    {
        $write = (bool)($params['write'] ?? false);
        $filename = $params['filename'] ?? null;
        $code     = $params['code'] ?? null;

        if (!$write || !$filename) {
            return $this;
        }

        $filename = $this->getOutputDir() . '/' . $filename;

        $parts = explode('/', substr($filename, 1));
        array_pop($parts);
        $dir = '';
        foreach ($parts as $part) {
            if (!is_dir(($dir .= "/$part"))) {
                mkdir($dir);
                chmod($dir, 0777);
            }
        }

        file_put_contents($filename, $code);
        chmod($filename, 0777);

        return $this;
    }
}
